# -*- coding: utf-8 -*-
"""
Monte Carlo simulation of columnar metal-carbon films: resistivity and strain sensitivity.

Generates hexagonal arrangements of columns (particles) of varying size, then calculates the total resistance using
a resistance law. The resistance in unstrained and strained state is used to find gauge factors.

@author: Silvan Schwebke
"""

import os.path
import pp
import sys
import math
from numpy import *
from time import time, strftime, localtime
from resistance import calculate_resistance_justsparse
from MCS_functions import resistance_total, create_grid,\
    MCS_find_neighbors, MCS_fix_diameters,\
    calc_accuracy, calc_separation_distribution,\
    mole_to_volume_fraction, elastic_modulus
#from MCS_visualizations import MCS_call_model
from MCS_job import MCS_job_full_hexagonal_varstrain
from MCS_plot_eval import k_plot_volfrac, k_plot_molefrac, k_plot_dist,\
    k_plot_deformation, R_plot_deformation, k_ratio_plot_molefrac,\
    k_ratio_plot_deformation, R_plot_molefrac
import matplotlib.pyplot as plt

versioninfo = 'MCS 0.52 published'

"""
=== Parameters ===
"""

save_to_file = True         # Save results to csv files
make_plot = True            # Create plots of the results
visualize = False           # Create a visualization of the column arrangement? If yes, only individual,
                            # non-parallel calculations are done to create pictures.
make_accuracy_plot = False  # Show plot of continuous accuracy estimation

# Material choice, to determine the applied elastic moduli
# E values are defined in the function elastic_modulus(), see MCS_functions.py.
# Options: Al2O3 (380 GPa), Pt (168 GPa), BN (80 GPa), Ni (200 GPa), C (80 GPa)
material_metal  = 'Ni'
material_matrix = 'C'

deformation = [1.002]  # Deformation applied for gauge factor calculation. Array of a single or
                       # multiple values.
                       
# Parameters of resistance calculation: R_total = R_grain + a * dist || b * exp(c*dist)
a = 1e3
b = 1e-9
c = 11.

d = 15.       # Mean column diameter
sigma = 0.35  # Standard deviation of column diameter

# (Densest packing of circles has a filling ratio of 0,907.)
mole_fraction_array = arange(0.55,0.75,0.02)

# Aspect ratio of the rectangle/cuboid used for the particle arrangement. Its absolute size is
# automatically scaled to achieve the desired metal content.
ax = 100.0
ay = 100.0
az = 0.0

n_particles = 100  # Number of particles, recommended value for low small-size effects: 1e4

two_dimensions = True  # 2D geometry (columns)

if two_dimensions == True:
    az = 0.0
n = int(n_particles)
# Number of nearest neighbors to consider:
if two_dimensions == True:
    n_neighbors = 6
else:
    n_neighbors = 12

(E_metal, E_matrix) = elastic_modulus(material_metal,material_matrix)
E_ratio = E_matrix/E_metal
n_deformation = len(deformation)
volume_fraction_array = mole_to_volume_fraction(mole_fraction_array, material_metal, material_matrix)

# Stop criteria
error_goal_k = 0.1            # Goal of relative uncertainty of kL and kT
nloops = 128                   # Maximum number of loops (choose large enough value here so array sizes suffice)
break_loop_time = 60.*60.*0.01 # Maximum time for each step in mole_fraction_array before calculation is aborted, in seconds

# Output arrays
final_kL = []
final_kT = []
final_kL_restotal = []
final_kT_restotal = []
final_restotal_R0 = []
final_restotal_RL = []
final_restotal_RT = []
final_error_rel_kL_restotal_upper = []
final_error_rel_kL_restotal_lower = []
final_error_rel_kT_restotal_upper = []
final_error_rel_kT_restotal_lower = []
final_error_rel_R0_restotal_upper = []
final_error_rel_R0_restotal_lower = []
final_error_rel_RL_restotal_upper = []
final_error_rel_RL_restotal_lower = []
final_error_rel_RT_restotal_upper = []
final_error_rel_RT_restotal_lower = []
final_list_dist_avg = []

if save_to_file==True:
    foldername = 'MCS_volfrac_' + strftime("%Y%m%d_%H%M%S", localtime())
    if not os.path.exists(foldername):
        os.makedirs(foldername)
        os.makedirs(foldername+'/otherplotformats')
        
    parameter_string = 'Simulation version: ' + versioninfo + '\n' \
        + 'a {:.2e}\n'.format(a) + 'b {:.2e}\n'.format(b) + 'c {:.2f}'.format(c) \
        + '\nMetal: ' + material_metal + ' (E={0} GPa)'.format(E_metal) \
        + '\nMatrix: ' + material_matrix + ' (E={0} GPa)'.format(E_matrix) \
        + '\nE_ratio ' + str(E_ratio) + '\nd ' + str(d) + '\nsigma ' + str(sigma) \
        + '\ndeformation ' + str(deformation) + '\n' + 'n ' + str(n) + '\n' \
        + 'n_neighbors ' + str(n_neighbors) + '\nmole_fraction ' + str(mole_fraction_array) + '\n'
    with open(foldername + '/parameters.dat', 'w') as file_:
        file_.write(parameter_string)
        
    results_string = 'molefrac deformation R0 RL RT kL_restot kT_restot u_kL_restot_upper u_kL_restot_lower u_kT_restot_upper u_kT_restot_lower n_loops timeelapsed\n'
    with open(foldername + '/results.dat', 'w') as file_:
        file_.write(results_string)

# tuple of all parallel python servers to connect with
ppservers = ()
#ppservers = ("127.0.0.1:60000", )
if visualize==True:
    job_server = pp.Server(1, ppservers=ppservers)
    print("Results visualization, therefore do individual (non-parallel) calculation and display results after step.")
else:
    if len(sys.argv) > 1:
        ncpus = int(sys.argv[1])
        # Creates jobserver with ncpus workers
        job_server = pp.Server(ncpus, ppservers=ppservers)
    else:
        # Creates jobserver with automatically detected number of workers
        job_server = pp.Server(ppservers=ppservers)
    print("Starting calculation with "+str(job_server.get_ncpus())+\
        " threads in parallel. Results are displayed as running means.")

n_parallel = job_server.get_ncpus()

# For a hexagonal grid we need to integrate the orientation over phi=0...30°.
# We are doing that in steps of 1.875°, i.e., do calculation in blocks of 16 loops.
nloops = 16*nloops

array_list_dist0 = {}
array_list_distL = {}
array_list_distT = {}

remaining_time_last_updated = time()

for fraction_idx in range(len(volume_fraction_array)):
    print 'Calculating ' + str(fraction_idx+1) + '/' + str(len(volume_fraction_array))
    avg_interval_time = 0.
    volume_fraction = volume_fraction_array[fraction_idx]
    mole_fraction = mole_fraction_array[fraction_idx]
    
    # Adjust the mole-fraction dependant diameter:
    #d = 1.762*math.exp(3.466*mole_fraction)
    #sigma = 0.02*d
    
    if two_dimensions == True:
        hex_spacing = sqrt( pi / (2.*sqrt(3.)*volume_fraction) ) * d
        s_wall = hex_spacing - d
        ra = ax/ay
        ax = sqrt( (ra * float(n) * pi/4. * d * d) / volume_fraction )
        ay = ax/ra
    else:
        hex_spacing = ( pi / (3.*sqrt(2.)*volume_fraction) )**(1./3.) * d
        s_wall = hex_spacing - d
        linear_scale_factor = ( ( 1./6. * pi * d**3. * n ) / ((ax)*(ay)*(az)) / volume_fraction)**(1./3.)
        ax = ax*linear_scale_factor
        ay = ay*linear_scale_factor
        az = az*linear_scale_factor
    
    array_restotal = zeros((nloops,n_deformation,3))
    array_kLkT_restotal = zeros((nloops,n_deformation,2))
    
    stat_rel_uncertainty_kLkT = ("inf","inf")
    j = 0
    loop_start_time = time()
    hex_angle = 0.0
    hex_angle_start = 0.0
    hex_angle_resolution = 16.0
    
    array_list_dist0[volume_fraction] = []
    array_list_distL[volume_fraction] = []
    array_list_distT[volume_fraction] = []
    
    while j<nloops:
        start_time = time()
        if visualize==True:
            array_list_dist = []
            hex_angle = 0.
            n, coordx, coordy, coordz, diameters_fixed, clusters0, clustersL, clustersT, job_restotal, job_kLkT_restotal, list_dist = MCS_job_full_hexagonal_varstrain(n,n_neighbors,ax,ay,az,d,sigma,hex_spacing,hex_angle,a,b,c,deformation,two_dimensions,visualize,E_ratio)
            if j<len(array_restotal):
                array_restotal[j,:,:] = job_restotal
                array_kLkT_restotal[j,:,:] = job_kLkT_restotal
                array_list_dist = hstack((array_list_dist,list_dist))
            else:
                array_restotal = vstack((array_restotal,job_restotal))
                array_kLkT_restotal = vstack((array_kLkT_restotal,job_kLkT_restotal))
                array_list_dist = vstack((array_list_dist,list_dist))
            j += 1
        else:
            inputs = range(n_parallel)
            jobs = [(input, job_server.submit(MCS_job_full_hexagonal_varstrain, (n,n_neighbors,ax,ay,az,d,sigma,hex_spacing,hex_angle_start+float(input)*30.0/hex_angle_resolution,a,b,c,deformation,two_dimensions,visualize,E_ratio),
                            (create_grid,MCS_find_neighbors,MCS_fix_diameters,resistance_total,calculate_resistance_justsparse,calc_separation_distribution),
                            ("numpy","MCS_functions","MCS_job","resistance","random","matplotlib"))) for input in inputs]
            for input, job in jobs:
                n, job_restotal, job_kLkT_restotal, job_list_dist0, job_list_distL, job_list_distT = job()
                if j<len(array_restotal):
                    array_restotal[j,:,:] = job_restotal
                    array_kLkT_restotal[j,:,:] = job_kLkT_restotal
                    array_list_dist0[volume_fraction] = hstack((array_list_dist0[volume_fraction],job_list_dist0))
                    array_list_distL[volume_fraction] = hstack((array_list_distL[volume_fraction],job_list_distL))
                    array_list_distT[volume_fraction] = hstack((array_list_distT[volume_fraction],job_list_distT))
                else:
                    array_restotal = vstack((array_restotal,job_restotal))
                    array_kLkT_restotal = vstack((array_kLkT_restotal,job_kLkT_restotal))
                    array_list_dist0[volume_fraction] = hstack((array_list_dist0[volume_fraction],job_list_dist0))
                    array_list_distL[volume_fraction] = hstack((array_list_distL[volume_fraction],job_list_distL))
                    array_list_distT[volume_fraction] = hstack((array_list_distT[volume_fraction],job_list_distT))
                j += 1
            
            hex_angle_start += float(n_parallel)*30.0/hex_angle_resolution
            if hex_angle_start>=30.0:
                hex_angle_start = 0
        
        avg_kLkT_restotal = mean(array_kLkT_restotal[:j,:,:], axis=0)
        avg_restotal = mean(array_restotal[:j,:,:], axis=0)

        kL_restotal = avg_kLkT_restotal[:,0]
        kT_restotal = avg_kLkT_restotal[:,1]
        
        interval_time = time()-start_time
        number_of_runs = j/n_parallel
        avg_interval_time = ((number_of_runs-1.)*avg_interval_time+interval_time) / float(number_of_runs)
        print('{}, R: {:.2e}, {:.2e}, {:.2e}, k_restot: {:.1f}, {:.1f}, t: {:.2f}s'.format(j,avg_restotal[0,0],avg_restotal[0,1],avg_restotal[0,2],kL_restotal[0],kT_restotal[0],interval_time))
        remaining_loops = len(volume_fraction_array) - fraction_idx - 1
        remaining_in_loop = (ceil(break_loop_time/avg_interval_time) - number_of_runs) * avg_interval_time
        remaining_time = remaining_in_loop + ceil(break_loop_time/avg_interval_time)*avg_interval_time * remaining_loops
        if time()-remaining_time_last_updated > 5*60.:
            remaining_time_last_updated = time()
            if remaining_time<3600.:
                print("      ###  Total time remaining approx. "+"{:1.1f}".format(remaining_time/60.)+" min  ###")
            else:
                print("      ###  Total time remaining approx. "+"{:1.1f}".format(remaining_time/3600.)+" h  ###")
        
        if j==16:
            error_max_restotal = 0.0
            for i in range(n_deformation):
                error_max_restotal = max(error_max_restotal, calc_accuracy(array_kLkT_restotal[:,i,:],j-1, error_goal_k, samples=5000, print_result=True, max_only=True)[0][0])
        if j%32==0:
            calcstats = job_server.get_stats()
            error_max_restotal = 0.0
            for i in range(n_deformation):
                error_max_restotal = max(error_max_restotal, calc_accuracy(array_kLkT_restotal[:,i,:],j-1, error_goal_k, samples=5000, print_result=True, max_only=True)[0][0])
            if error_max_restotal<error_goal_k:
                break

        if visualize==True:
            #MCS_call_model(deformation,ax,ay,az,coordx,coordy,coordz,diameters_fixed,clusters0,clustersL,clustersT)
            if j>1:
                break
        
        if j%hex_angle_resolution==0 and time()-loop_start_time > break_loop_time:
            for i in range(n_deformation):
                error_max_restotal = max(error_max_restotal, calc_accuracy(array_kLkT_restotal[:,i,:],j-1, error_goal_k, samples=5000, print_result=True, max_only=True)[0][0])
            print('Exceeded maximum time ({:.1f} min), current step is being aborted.'.format(break_loop_time/60.))
            break

    if visualize==False:
        for i in range(n_deformation):
            error_kL_restotal, error_rel_kL_restotal, error_kT_restotal, error_rel_kT_restotal, error_max_restotal = calc_accuracy(array_kLkT_restotal[:,i,:],j-1, error_goal_k, samples=5000, print_result=False)
            error_R0_restotal, error_rel_R0_restotal, error_max_R0_restotal = calc_accuracy(array_restotal[:,i,0],j-1, error_goal_k, samples=5000, print_result=False)
            error_RL_restotal, error_rel_RL_restotal, error_RT_restotal, error_rel_RT_restotal, error_max_R_restotal = calc_accuracy(array_restotal[:,i][:,1:3],j-1, error_goal_k, samples=5000, print_result=False)
        
            final_kL_restotal = hstack((final_kL_restotal,avg_kLkT_restotal[i,0]))
            final_kT_restotal = hstack((final_kT_restotal,avg_kLkT_restotal[i,1]))
            final_restotal_R0 = hstack((final_restotal_R0,avg_restotal[i,0]))
            final_restotal_RL = hstack((final_restotal_RL,avg_restotal[i,1]))
            final_restotal_RT = hstack((final_restotal_RT,avg_restotal[i,2]))
            final_error_rel_kL_restotal_upper = hstack((final_error_rel_kL_restotal_upper,error_rel_kL_restotal[1]))
            final_error_rel_kL_restotal_lower = hstack((final_error_rel_kL_restotal_lower,error_rel_kL_restotal[0]))
            final_error_rel_kT_restotal_upper = hstack((final_error_rel_kT_restotal_upper,error_rel_kT_restotal[1]))
            final_error_rel_kT_restotal_lower = hstack((final_error_rel_kT_restotal_lower,error_rel_kT_restotal[0]))
            final_error_rel_R0_restotal_upper = hstack((final_error_rel_R0_restotal_upper,error_rel_R0_restotal[1]))
            final_error_rel_R0_restotal_lower = hstack((final_error_rel_R0_restotal_lower,error_rel_R0_restotal[0]))
            final_error_rel_RL_restotal_upper = hstack((final_error_rel_RL_restotal_upper,error_rel_RL_restotal[1]))
            final_error_rel_RL_restotal_lower = hstack((final_error_rel_RL_restotal_lower,error_rel_RL_restotal[0]))
            final_error_rel_RT_restotal_upper = hstack((final_error_rel_RT_restotal_upper,error_rel_RT_restotal[1]))
            final_error_rel_RT_restotal_lower = hstack((final_error_rel_RT_restotal_lower,error_rel_RT_restotal[0]))
        
            if save_to_file==True:
                results_string = str(mole_fraction) + ' ' + str(deformation[i]) \
                    + ' ' + str(avg_restotal[i,0]) + ' ' + str(avg_restotal[i,1]) + ' ' + str(avg_restotal[i,2]) \
                    + ' ' + str(avg_kLkT_restotal[i,0]) + ' ' + str(avg_kLkT_restotal[i,1]) \
                    + ' ' + str(error_kL_restotal[1]) + ' ' + str(error_kL_restotal[0]) \
                    + ' ' + str(error_kT_restotal[1]) + ' ' + str(error_kT_restotal[0]) \
                    + ' ' + str(j) + ' ' + str(time()-loop_start_time) + '\n'
                with open(foldername + '/results.dat', 'a') as file_:
                    file_.write(results_string)
            
    job_server.print_stats()

job_server.destroy()

if make_plot==True:
    #k_plot_volfrac(final_kL_restotal,final_kT_restotal,vstack((final_error_rel_kL_restotal_lower,final_error_rel_kL_restotal_upper)),vstack((final_error_rel_kT_restotal_lower,final_error_rel_kT_restotal_upper)),volume_fraction_array)
    #k_plot_volfrac(final_kL_restotal,final_kT_restotal,vstack((final_error_rel_kL_restotal_lower,final_error_rel_kL_restotal_upper)),vstack((final_error_rel_kT_restotal_lower,final_error_rel_kT_restotal_upper)),volume_fraction_array)
    if save_to_file==False:
        foldername = False
    v = len(mole_fraction_array)
    w = len(deformation)
    if w>1:
        for i in range(v):
            k_plot_deformation(final_kL_restotal[i*w:(i+1)*w],final_kT_restotal[i*w:(i+1)*w],vstack((final_error_rel_kL_restotal_lower[i*w:(i+1)*w],final_error_rel_kL_restotal_upper[i*w:(i+1)*w])),vstack((final_error_rel_kT_restotal_lower[i*w:(i+1)*w],final_error_rel_kT_restotal_upper[i*w:(i+1)*w])),deformation,save_to_folder=foldername)
            k_ratio_plot_deformation(final_kL_restotal[i*w:(i+1)*w],final_kT_restotal[i*w:(i+1)*w],vstack((final_error_rel_kL_restotal_lower[i*w:(i+1)*w],final_error_rel_kL_restotal_upper[i*w:(i+1)*w])),vstack((final_error_rel_kT_restotal_lower[i*w:(i+1)*w],final_error_rel_kT_restotal_upper[i*w:(i+1)*w])),deformation,save_to_folder=foldername)
            #R_plot_deformation(final_restotal_RL[i*w:(i+1)*w],final_restotal_RT[i*w:(i+1)*w],vstack((final_error_rel_RL_restotal_lower[i*w:(i+1)*w],final_error_rel_RL_restotal_upper[i*w:(i+1)*w])),vstack((final_error_rel_RT_restotal_lower[i*w:(i+1)*w],final_error_rel_RT_restotal_upper[i*w:(i+1)*w])),deformation)
    if v>1:
        for i in range(w):
            final_list_dist_avg = []
            for volidx in volume_fraction_array:
                final_list_dist_avg = hstack((final_list_dist_avg,mean(array_list_dist0[volidx])))
            parameter_string = 'Parameters: n={0}'.format(n_particles) + ', a={:.2e}'.format(a) \
                + ', b={:.2e}'.format(b) + ', c={:.1f}, '.format(c) \
                + '$\mathrm{\mathsf{\epsilon=}}$' + '{0}, '.format(deformation[i]-1) \
                + '$\mathrm{\mathsf{E_{Matrix}/E_{Metal}=}}$' + '{:.2f}'.format(E_ratio)
            k_plot_molefrac(final_kL_restotal[i::w],final_kT_restotal[i::w],vstack((final_error_rel_kL_restotal_lower[i::w],final_error_rel_kL_restotal_upper[i::w])),vstack((final_error_rel_kT_restotal_lower[i::w],final_error_rel_kT_restotal_upper[i::w])),mole_fraction_array,parameter_string,save_to_folder=foldername)
            k_ratio_plot_molefrac(final_kL_restotal[i::w],final_kT_restotal[i::w],vstack((final_error_rel_kL_restotal_lower[i::w],final_error_rel_kL_restotal_upper[i::w])),vstack((final_error_rel_kT_restotal_lower[i::w],final_error_rel_kT_restotal_upper[i::w])),mole_fraction_array,parameter_string,save_to_folder=foldername)
            k_plot_dist(final_kL_restotal[i::w],final_kT_restotal[i::w],vstack((final_error_rel_kL_restotal_lower[i::w],final_error_rel_kL_restotal_upper[i::w])),vstack((final_error_rel_kT_restotal_lower[i::w],final_error_rel_kT_restotal_upper[i::w])),final_list_dist_avg,"($\mathrm{\mathsf{\epsilon=}}$"+str(deformation[i]-1)+")",save_to_folder=foldername)
            R_plot_molefrac(final_restotal_R0[i::w],vstack((final_error_rel_R0_restotal_lower[i::w],final_error_rel_R0_restotal_upper[i::w])),mole_fraction_array,parameter_string,save_to_folder=foldername)
