# Source: https://gist.github.com/aelguindy/1747940
# see also: http://math.stackexchange.com/questions/106062/algorithm-to-find-the-total-resistance-between-2-vertices-of-a-graph#106071

# SOURCE: http://elonen.iki.fi/code/misc-notes/python-gaussj/

import numpy
import scipy.sparse.linalg

def gauss_jordan(m, eps = 1.0/(10**300)):
  """Puts given matrix (2D array) into the Reduced Row Echelon Form.
     Returns True if successful, False if 'm' is singular.
     NOTE: make sure all the matrix items support fractions! Int matrix will NOT work!
     Written by Jarno Elonen in April 2005, released into Public Domain"""
  (h, w) = (len(m), len(m[0]))
  for y in range(0,h):
    maxrow = y
    for y2 in range(y+1, h):    # Find max pivot
      if abs(m[y2][y]) > abs(m[maxrow][y]):
        maxrow = y2
    (m[y], m[maxrow]) = (m[maxrow], m[y])
    if abs(m[y][y]) <= eps:     # Singular?
      return False
    for y2 in range(y+1, h):    # Eliminate column y
      c = m[y2][y] / m[y][y]
      for x in range(y, w):
        m[y2][x] -= m[y][x] * c
  for y in range(h-1, 0-1, -1): # Backsubstitute
    c  = m[y][y]
    for y2 in range(0,y):
      for x in range(w-1, y-1, -1):
        m[y2][x] -=  m[y][x] * m[y2][y] / c
    m[y][y] /= c
    for x in range(h, w):       # Normalize row y
      m[y][x] /= c
  return True

def preprocess(edges, V):
    lists = []
    for i in range(V): lists.append([])
    for fro, to, res in edges:
        lists[fro].append((to, res))
        lists[to].append((fro, res))
    return lists

def make_eqns(lists, node1, node2):
    coeffs = []
    Vars = len(lists)
    for i, l in enumerate(lists):
        cs = [0.0]*Vars
        cs[i] = sum(1/b for (a, b) in l)
        for other, res in l:
            cs[other] -= 1.0/res
        coeffs.append(cs)
    rhs = [0] * Vars
    rhs[node1] = 1.0
    rhs[node2] = -1.0
    n = max(node1, node2)
    coeffs = [c[:n] + c[n + 1:] for c in coeffs]
    return coeffs[:-1], rhs[:-1]

def make_eqns_sparse(lists, node1, node2):
    # Attention: Algorithm works only if there is no more than one resistance between each two nodes.
    row = []
    col = []
    data = []
    Vars = len(lists)
    n = max(node1, node2)
    for i, l in enumerate(lists):
        if i < Vars-1:
            #cs = [0.0]*Vars
            #cs[i] = sum(1/b for (a, b) in l)
            value = sum(1/b for (a, b) in l)
            data.append(value)
            row.append(i)
            col.append(i)
            for other, res in l:
                #cs[other] -= 1.0/res
                if other != n:
                    othervalue = -1.0/res
                    data.append(othervalue)
                    row.append(i)
                    col.append(other)
    rhs = [0] * Vars
    rhs[node1] = 1.0
    rhs[node2] = -1.0
    coeffs_sparse = scipy.sparse.csr_matrix((data, (row, col)), shape=(Vars-1, Vars-1))
    return coeffs_sparse, numpy.asarray(rhs[:-1])

def calculate_resistance(nodes, edges, source, destination):
    src = source
    dst = destination
    ls = preprocess(edges, nodes)
    a, b = make_eqns(ls, src, dst)
    M = [a[i] + [b[i]] for i in range(len(a))]
    gauss_jordan(M)
    return abs(M[min(src, dst)][-1])

def calculate_resistance_numpy(nodes, edges, source, destination):
    src = source
    dst = destination
    ls = preprocess(edges, nodes)
    a, b = make_eqns(ls, src, dst)
    x = numpy.linalg.solve(a, b)
    return abs(x[min(src, dst)])

def calculate_resistance_sparse(nodes, edges, source, destination):
    src = source
    dst = destination
    ls = preprocess(edges, nodes)
    a, b = make_eqns(ls, src, dst)
    sa = scipy.sparse.csr_matrix(a)
    x = scipy.sparse.linalg.spsolve(sa, b)
    if scipy.sparse.issparse(x):
        return abs(x.toarray()[min(src, dst)])
    else:
        return abs(x[min(src, dst)])

def calculate_resistance_justsparse(nodes, edges, source, destination):
    src = source
    dst = destination
    ls = preprocess(edges, nodes)
    sa, b = make_eqns_sparse(ls, src, dst)
    x = scipy.sparse.linalg.spsolve(sa, b)
    if scipy.sparse.issparse(x):
        return abs(x.toarray()[min(src, dst)])
    else:
        return abs(x[min(src, dst)])

def calculate_resistance_approx(nodes, edges, source, destination):
    src = source
    dst = destination
    ls = preprocess(edges, nodes)
    a, b = make_eqns(ls, src, dst)
    x, residuals, rank, s = numpy.linalg.lstsq(a, b)
    return abs(x[min(src, dst)])

if __name__ == '__main__':
    print 'Self test:'
    print 'a: 0.5?'
    print str(calculate_resistance(4, [(0, 1, 1.0), (0, 2, 1.0), (0, 3, 1.0), (1, 3, 1.0), (2, 3, 1.0)], 0, 3)) + ' (calculate_resistance)'
    print str(calculate_resistance_numpy(4, [(0, 1, 1.0), (0, 2, 1.0), (0, 3, 1.0), (1, 3, 1.0), (2, 3, 1.0)], 0, 3)) + ' (calculate_resistance_numpy)'
    print str(calculate_resistance_sparse(4, [(0, 1, 1.0), (0, 2, 1.0), (0, 3, 1.0), (1, 3, 1.0), (2, 3, 1.0)], 0, 3)) + ' (calculate_resistance_sparse)'
    print str(calculate_resistance_approx(4, [(0, 1, 1.0), (0, 2, 1.0), (0, 3, 1.0), (1, 3, 1.0), (2, 3, 1.0)], 0, 3)) + ' (calculate_resistance_approx)'
    print str(calculate_resistance_justsparse(4, [(0, 1, 1.0), (0, 2, 1.0), (0, 3, 1.0), (1, 3, 1.0), (2, 3, 1.0)], 0, 3)) + ' (calculate_resistance_justsparse)'
    print 'b: 0.667?'
    print str(calculate_resistance(3, [(0, 1, 1.0), (0, 2, 1.0), (1, 2, 1.0)], 0, 2)) + ' (calculate_resistance)'
    print str(calculate_resistance_numpy(3, [(0, 1, 1.0), (0, 2, 1.0), (1, 2, 1.0)], 0, 2)) + ' (calculate_resistance_numpy)'
    print str(calculate_resistance_sparse(3, [(0, 1, 1.0), (0, 2, 1.0), (1, 2, 1.0)], 0, 2)) + ' (calculate_resistance_sparse)'
    print str(calculate_resistance_approx(3, [(0, 1, 1.0), (0, 2, 1.0), (1, 2, 1.0)], 0, 2)) + ' (calculate_resistance_approx)'
    print str(calculate_resistance_justsparse(3, [(0, 1, 1.0), (0, 2, 1.0), (1, 2, 1.0)], 0, 2)) + ' (calculate_resistance_justsparse)'
